#include <clang-c/Index.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "locale_syms.h"
#include "xlocale_syms.h"

static int containsErrors;
static CXTranslationUnit translationUnit;
static const char *functionName;

typedef struct stack
{
	struct stack *prev;
	CXCursor func;
	CXCursor call;
} stack;

/*
 * List of globals that reentrant code is allowed to reference.
 */
static const char* permittedGlobals[] = {
	"__xlocale_global_locale",
	"__isthreaded",
	"__stdinp",
	"__stdoutp",
	"__stderrp",
	"tzname",
	0
};

static int isInList(const char **list, const char *sym)
{
	// FIXME: This is slow, but I don't care.  Fix it if it actually matters to
	// overall performance.
	while (*list)
	{
		if (strcmp(sym, *list) == 0)
		{
			return 1;
		}
		list++;
	}
	return 0;
}

static void freestring(CXString *str)
{
	clang_disposeString(*str);
}
#define SCOPED_STR(name, value)\
		__attribute__((unused))\
	__attribute__((cleanup(freestring))) CXString name ## str = value;\
	const char *name = clang_getCString(name ## str);

int isGlobal(CXCursor cursor)
{
	// If something has any linkage type, it is definitely persistent
	if (clang_getCursorLinkage(cursor) != CXLinkage_NoLinkage)
	{
		return 1;
	}
	CXSourceRange r = clang_getCursorExtent(cursor);
	CXToken *tokens;
	unsigned numTokens;
	clang_tokenize(translationUnit, r, &tokens, &numTokens);
	int ret = 0;
	for (int i=0 ; i<numTokens ; i++)
	{
		if (clang_getTokenKind(tokens[i]) == CXToken_Keyword)
		{
			SCOPED_STR(tok, clang_getTokenSpelling(translationUnit, tokens[i]));
			if (strcmp(tok, "static") == 0)
			{
				ret = 1;
				break;
			}
		}
	}
	clang_disposeTokens(translationUnit, tokens, numTokens);
	return ret;
}

void printCursorLocation(CXCursor cursor, int indent)
{
	CXSourceRange r = clang_getCursorExtent(cursor);
	CXSourceLocation l = clang_getRangeStart(r);
	CXFile file;
	unsigned line, column;
	clang_getSpellingLocation(l, &file, &line, &column, 0);
	CXString fileName = clang_getFileName(file);
	fprintf(stderr, "\n%s%s:%u:%u ", indent ? "\t" : "",
			clang_getCString(fileName), line, column);
	clang_disposeString(fileName);
}

void printStack(stack *s)
{
	int indent = 0;
	while (s)
	{
		indent = 1;
		printCursorLocation(s->call, 1);
		s = s->prev;
	}
}
int checkRecursion(stack *s, CXCursor c)
{
	while (s)
	{
		if (clang_equalCursors(s->func, c))
		{
			return 1;
		}
		s = s->prev;
	}
	return 0;
}

enum CXChildVisitResult visitFunction(CXCursor cursor, CXCursor parent,
		CXClientData client_data)
{
	if (cursor.kind == CXCursor_DeclRefExpr)
	{
		CXCursor ref = clang_getCursorReferenced(cursor);
		if ((ref.kind == CXCursor_VarDecl) &&
			!clang_isConstQualifiedType(clang_getCanonicalType(clang_getCursorType(ref))) && 
			isGlobal(ref)) 
		{
			SCOPED_STR(staticref, clang_getCursorSpelling(ref));
			if (!isInList(permittedGlobals, staticref))
			{
				printCursorLocation(cursor, 0);
				fprintf(stderr, " references to to global state via %s from:",
					staticref);
				printStack(client_data);
				containsErrors = 1;
			}
		}
		else if (ref.kind == CXCursor_FunctionDecl)
		{
			SCOPED_STR(func, clang_getCursorSpelling(ref));
			if (isInList(locale_syms, func))
			{
				printCursorLocation(cursor, 0);
				fprintf(stderr, " calls thread-unsafe function %s() from:",
						func);
				printStack(client_data);
				containsErrors = 1;
			}
			else if (!checkRecursion(client_data, ref))
			{
				stack s = {client_data, ref, cursor};
				clang_visitChildren(ref, visitFunction, &s);
			}
		}
	}
	return CXChildVisit_Recurse;
}

enum CXChildVisitResult visitFunctions(CXCursor cursor, CXCursor parent,
		CXClientData client_data)
{
	if (cursor.kind == CXCursor_FunctionDecl &&
	    clang_isCursorDefinition(cursor))
	{
		SCOPED_STR(fname, clang_getCursorSpelling(cursor));
		if (isInList(xlocale_syms, fname))
		{
			stack s = {0, cursor, cursor};
			clang_visitChildren(cursor, visitFunction, &s);
		}
	}
	return CXChildVisit_Recurse;
}


int main(int argc, char **argv)
{
	CXIndex idx = clang_createIndex(1, 0);
	translationUnit =
		clang_createTranslationUnitFromSourceFile(idx, 0, argc, (void*)argv, 0, 0);
	if (translationUnit)
	{
		clang_visitChildren(clang_getTranslationUnitCursor(translationUnit),
				visitFunctions, 0);
		// If we have errors, abort, otherwise invoke CC
		if (containsErrors)
		{
			return 1;
		}
	}
	char **argvn = calloc(argc+1, sizeof(char*));
	memcpy(argvn, argv, argc*sizeof(char*));
	argvn[0] = "/usr/bin/cc";
	execv(argvn[0], argvn);
	perror("Failed to invoke cc");
	return 1;
}
